// Device-Independent Bitmap manipulation
// (C) W.A.Steer 1996

// this version modified 9/96

// Y-offset lookup now allocated dynamically


// To do:

//  add a option to open file for bitmap info only, without bits

#include <owl.h>
#include <io.h>
#include <alloc.h>

#include "..\fastcadr\dibobj17.h"


TDIBobject::TDIBobject(LPSTR filename, UINT &errorval, UINT flags)
{
 errorval=DOE_SUCCESS;

 long ioresult;

 BITMAPFILEHEADER bfh;
 HFILE theFile; WORD sizeBI;

 WORD sizeBFH=sizeof(BITMAPFILEHEADER);

 theFile=_lopen(filename, READ);


 if (theFile!=HFILE_ERROR)
 {
  ioresult=_lread(theFile, &bfh, sizeBFH);

  if (ioresult<sizeBFH) errorval=DOE_UXEOF;
  if (ioresult==HFILE_ERROR) errorval=DOE_READFAILED;

  if (errorval==DOE_SUCCESS)
  {
   if ((bfh.bfType!=0x4D42) || (bfh.bfOffBits<sizeBFH))
   {
    errorval=DOE_INVALIDBMP;
   }
  }


  if (errorval==DOE_SUCCESS)
  {
   sizeBI=bfh.bfOffBits-sizeBFH;

   pbmi=(BITMAPINFO *) new char[sizeBI];

   if (pbmi==NULL) errorval=DOE_OUTOFMEM;
  }

  if (errorval==DOE_SUCCESS)
  {
   ioresult=_lread(theFile, (LPSTR)pbmi, sizeBI);

   if (ioresult<sizeBI) errorval=DOE_UXEOF;
   if (ioresult==HFILE_ERROR) errorval=DOE_READFAILED;
  }

  if (errorval==DOE_SUCCESS)
  {
   if (pbmi->bmiHeader.biSize != sizeof(BITMAPINFOHEADER))
        errorval=DOE_INVALIDBMP;

   if (pbmi->bmiHeader.biSize == sizeof(BITMAPCOREHEADER))
        errorval=DOE_OS2BMP;
  }

  if ((errorval==DOE_SUCCESS) && (flags & DO_RASTER))
  {
   if (pbmi->bmiHeader.biCompression!=BI_RGB)
        errorval=DOE_COMPRESSED;
   if (pbmi->bmiHeader.biBitCount!=8)
        errorval=DOE_NOT8BIT;
  }


  if (!(flags & DO_HEADERONLY))
  {
   if (errorval==DOE_SUCCESS)
   {
    if (pbmi->bmiHeader.biClrUsed==0)
         pbmi->bmiHeader.biClrUsed=1L << pbmi->bmiHeader.biBitCount;

    if (pbmi->bmiHeader.biSizeImage==0) pbmi->bmiHeader.biSizeImage=
         filelength(theFile)-bfh.bfOffBits;

    bytesLine=((pbmi->bmiHeader.biWidth+3)/4)*4;

    GlobalCompact(-1);

    BitsHandle=(HBITMAP)GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT,
                         pbmi->bmiHeader.biSizeImage  +bytesLine);
//                                                    ^^^^^^^^^^^
//  simple HACK to fixup problem with ADOBEPS.DRV print driver

    if (BitsHandle==NULL) errorval=DOE_OUTOFMEM;
   }


   if (errorval==DOE_SUCCESS)
   {
    LPSTR BitsPtr=(LPSTR)GlobalLock(BitsHandle);

    ioresult=_hread(theFile, BitsPtr, pbmi->bmiHeader.biSizeImage);

    if (ioresult<pbmi->bmiHeader.biSizeImage) errorval=DOE_UXEOF;
    if (ioresult==HFILE_ERROR) errorval=DOE_READFAILED;

    GlobalUnlock(BitsHandle);
   }
  }
  else
  {
   BitsHandle=NULL;
  }

  _lclose(theFile);
 }
 else
 {
  errorval=DOE_OPENFAILED;
 }


 if (errorval==DOE_SUCCESS)
 {
  hpal=CreateHpal();
 }
 else
 {
  hpal=NULL;
 }
}



TDIBobject::~TDIBobject()
{
 delete[] pbmi;
 if (BitsHandle!=NULL) GlobalFree(BitsHandle);
 if (hpal!=NULL) DeleteObject(hpal);
}


void TDIBobject::Save(LPSTR Name, UINT &errorval)
{
 errorval=DOE_SUCCESS;

 long ioresult;

 BITMAPFILEHEADER bfh;

 HFILE theFile;


 theFile=_lcreat(Name,0);

 if (theFile!=HFILE_ERROR) 
 {
  bfh.bfOffBits=sizeof(BITMAPFILEHEADER)+
		sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*pbmi->bmiHeader.biClrUsed;

  bfh.bfReserved1=0;
  bfh.bfReserved2=0;

  bfh.bfSize=bfh.bfOffBits+pbmi->bmiHeader.biSizeImage;
  bfh.bfType=0x4D42;


  ioresult=_lwrite(theFile, &bfh, sizeof(BITMAPFILEHEADER));

  if (ioresult!=sizeof(BITMAPFILEHEADER)) errorval=DOE_WRITEFAILED;

  if (errorval==DOE_SUCCESS)
  {
   ioresult=_lwrite(theFile, (LPSTR)pbmi,
		 sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*(WORD)pbmi->bmiHeader.biClrUsed);
 
   if (ioresult!=(sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*(WORD)pbmi->bmiHeader.biClrUsed)) errorval=DOE_WRITEFAILED;
  }

  if (errorval==DOE_SUCCESS)
  {
   LPSTR BitsPtr=(LPSTR)GlobalLock(BitsHandle);

   ioresult=_hwrite(theFile, BitsPtr, pbmi->bmiHeader.biSizeImage);

   if (ioresult!=pbmi->bmiHeader.biSizeImage) errorval=DOE_WRITEFAILED;

   GlobalUnlock(BitsHandle);
  }

  _lclose(theFile);

  if (errorval!=DOE_SUCCESS) remove((char*)Name);
 }
 else
 {
  errorval=DOE_OPENFAILED;
 }
}



void TDIBobject::DIBtoDevice(HDC aDC, int XDest, int YDest, int cx, int cy, int XSrc, int YSrc)
{
 LPSTR BitsPtr = (LPSTR)GlobalLock(BitsHandle);

 SetDIBitsToDevice(aDC,XDest,YDest,cx,cy,XSrc,YSrc,
     0,(unsigned int)pbmi->bmiHeader.biHeight,BitsPtr,pbmi,DIB_RGB_COLORS);

 GlobalUnlock(BitsHandle);
}

void TDIBobject::StretchDIBtoDevice(HDC aDC, int XDest, int YDest, int cxDest, int cyDest, int XSrc, int YSrc, int cxSrc, int cySrc)
{
 LPSTR BitsPtr = (LPSTR)GlobalLock(BitsHandle);

 int oldSmode = SetStretchBltMode(aDC, STRETCH_DELETESCANS);

 StretchDIBits(aDC,XDest,YDest,cxDest,cyDest,XSrc,YSrc,cxSrc,cySrc,
     BitsPtr,pbmi,DIB_RGB_COLORS,SRCCOPY);

 SetStretchBltMode(aDC, oldSmode);

 GlobalUnlock(BitsHandle);
}

HBITMAP TDIBobject::CreateDDB(HDC aDC)
{
 LPSTR BitsPtr = (LPSTR)GlobalLock(BitsHandle);

 HBITMAP aDDbitmap=CreateDIBitmap(aDC, (BITMAPINFOHEADER FAR*) pbmi, CBM_INIT,
		BitsPtr, pbmi, DIB_RGB_COLORS);

 GlobalUnlock(BitsHandle);

 return aDDbitmap;
}


HPALETTE TDIBobject::CreateHpal()
{
 HPALETTE hpal;

 WORD numColors=pbmi->bmiHeader.biClrUsed;

 LPLOGPALETTE pal;
 pal=(LPLOGPALETTE) farmalloc(sizeof(LOGPALETTE)+sizeof(PALETTEENTRY)*numColors);

 pal->palVersion=0x300;
 pal->palNumEntries=numColors;

 for (int u=0; u<numColors; u++)
 {
  pal->palPalEntry[u].peRed=pbmi->bmiColors[u].rgbRed;
  pal->palPalEntry[u].peGreen=pbmi->bmiColors[u].rgbGreen;
  pal->palPalEntry[u].peBlue=pbmi->bmiColors[u].rgbBlue;
  pal->palPalEntry[u].peFlags=NULL;
 }

 hpal=CreatePalette(pal);
 farfree(pal);

 return hpal;
}



// Everything beneath here is for 8-bit bitmap manipulation purposes only

TDIBobjras::TDIBobjras(WORD width, WORD height, WORD numColors, RGBQUAD &colors, UINT &errorval)
{
 errorval=DOE_SUCCESS;

 BitsHandle=NULL;

 pbmi=(BITMAPINFO *) new char[sizeof(BITMAPINFOHEADER)+numColors*sizeof(RGBQUAD)];

 if (pbmi==NULL) errorval=DOE_OUTOFMEM;

 if (errorval==DOE_SUCCESS)
 {
  bytesLine=((width+3)/4)*4;

  pbmi->bmiHeader.biSize          = sizeof(BITMAPINFOHEADER);
  pbmi->bmiHeader.biWidth         = width;
  pbmi->bmiHeader.biHeight        = height;
  pbmi->bmiHeader.biPlanes        = 1; // always
  pbmi->bmiHeader.biBitCount      = 8;
  pbmi->bmiHeader.biCompression   = BI_RGB; // normally
  pbmi->bmiHeader.biSizeImage     = bytesLine*height;
  pbmi->bmiHeader.biXPelsPerMeter = 0;
  pbmi->bmiHeader.biYPelsPerMeter = 0;
  pbmi->bmiHeader.biClrUsed       = numColors;
  pbmi->bmiHeader.biClrImportant  = numColors;

  memcpy(pbmi->bmiColors, &colors, sizeof(RGBQUAD)*numColors);


  GlobalCompact(-1);

  BitsHandle=(HBITMAP)GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT,
                        pbmi->bmiHeader.biSizeImage  +bytesLine);
//                                                  ^^^^^^^^^^^
// simple HACK to fixup problem with ADOBEPS.DRV print driver

  if (BitsHandle==NULL) errorval=DOE_OUTOFMEM;
 }

 if (errorval==DOE_SUCCESS)
 {
  hpal=CreateHpal();
 }
 else
 {
  hpal=NULL;
 }
}



extern "C" {
void FAR PASCAL __ahIncr(); }


WORD TDIBobjras::GetPixelPaletteIndex(WORD x, WORD y)
{
 DWORD offs=x+y*bytesLine;

 long addr=MAKELONG(LOWORD(offs), HIWORD(bits) +
		      HIWORD(offs) * FP_OFF(__ahIncr));

 return *(BYTE FAR *)addr;
}


RGBQUAD TDIBobjras::GetPixelRGB(WORD x, WORD y)
{
 DWORD offs=x+y*bytesLine;

 long addr=MAKELONG(LOWORD(offs), HIWORD(bits) +
   		      HIWORD(offs) * FP_OFF(__ahIncr));

 return (pbmi->bmiColors[*(BYTE FAR *)addr]);
}


void TDIBobjras::SetPixelPaletteIndex(WORD x, WORD y, WORD PalIndex)
{
 DWORD offs=x+y*bytesLine;

 long addr=MAKELONG(LOWORD(offs), HIWORD(bits) +
		      HIWORD(offs) * FP_OFF(__ahIncr));

 *(BYTE FAR *)addr=BYTE(PalIndex);
}


void TDIBobjras::SetLine(int L, BYTE *pixels)
{
 DWORD offs=L*bytesLine;

 hmemcpy( MK_FP(HIWORD(bits)+HIWORD(offs)*FP_OFF(__ahIncr),
  LOWORD(offs)), pixels, pbmi->bmiHeader.biWidth);
}


void TDIBobjras::GetLine(int L, BYTE *pixels)
{
 DWORD offs=L*bytesLine;

 hmemcpy( pixels, MK_FP(HIWORD(bits)+HIWORD(offs)*FP_OFF(__ahIncr),
  LOWORD(offs)), pbmi->bmiHeader.biWidth);
}

/*

void TDIBobjras::Wipe()
{
 _fmemset((BYTE FAR *)bits,0,pbmi->bmiHeader.biWidth);

 long addr;

 for (int y=0; y<pbmi->bmiHeader.biHeight; y++)
 {
  addr=MAKELONG(LOWORD(offset[y]), HIWORD(bits) +
		      HIWORD(offset[y]) * FP_OFF(__ahIncr));

  hmemcpy((BYTE FAR *)addr, (BYTE FAR *)bits,pbmi->bmiHeader.biWidth);
 }
}


void TDIBobjras::LineDraw(int x1, int y1, int x2, int y2, WORD PalIndex)
{// kwikfix
 if ((x1>0) && (x2>0) && (y1>0) && (y2>0) &&
  (x1<pbmi->bmiHeader.biWidth ) && (x2<pbmi->bmiHeader.biWidth) &&
  (y1<pbmi->bmiHeader.biHeight) && (y2<pbmi->bmiHeader.biHeight))
{
 DWORD offs; long addr;
 register int c,d,xl,yl;

 xl=x2-x1; yl=y2-y1;
                            
 if (abs(yl)>abs(xl))
 {
  if (y2<y1) { c=y1; y1=y2; y2=c; c=x1; x1=x2; x2=c; }

  for (c=y1; c<=y2; c++)
  {
   d=MulDiv(c-y1,xl,yl)+x1;

//   SetPixelPaletteIndex(d,c,PalIndex);
   offs=d+offset[c];

   addr=MAKELONG(LOWORD(offs), HIWORD(bits) +
		      HIWORD(offs) * FP_OFF(__ahIncr));

   *(BYTE FAR *)addr=BYTE(PalIndex);
  }
  
 }
 else
 {
  if (x2<x1) { c=y1; y1=y2; y2=c; c=x1; x1=x2; x2=c; }

  for (c=x1; c<=x2; c++)
  {
   d=MulDiv(c-x1,yl,xl)+y1;

//   SetPixelPaletteIndex(c,d,PalIndex);
   offs=c+offset[d];

   addr=MAKELONG(LOWORD(offs), HIWORD(bits) +
		      HIWORD(offs) * FP_OFF(__ahIncr));

   *(BYTE FAR *)addr=BYTE(PalIndex);
  }
 }

}

}

void TDIBobjras::TriangleFill(POINT *corner, WORD PalIndex)
{
 int x,y;
 int maxy=-32767, maxp, miny=32767, minp, midp, minx=32767, maxx=-32767;
 for (int register p=0; p<3; p++)
 {
  x=corner[p].x; y=corner[p].y;

  if (y<miny) { miny=y; minp=p; }
  if (y>maxy) { maxy=y; maxp=p; }
  if (x<minx) minx=x;
  if (x>maxx) maxx=x;
 }

 if ((miny>=0) && (maxy<pbmi->bmiHeader.biHeight) &&
              (minx>=0) && (maxx<pbmi->bmiHeader.biWidth))
 {
  int tmp; int x1,x2;

  for (p=0; p<3; p++) { if ((p!=minp) && (p!=maxp)) midp=p; };

  int c0x=corner[minp].x, c0y=corner[minp].y;
  int c1x=corner[midp].x, c1y=corner[midp].y;
  int c2x=corner[maxp].x, c2y=corner[maxp].y;

  int xal=c2x-c0x;
  int yal=c2y-c0y;

  int xbl=c1x-c0x;
  int ybl=c1y-c0y;

  int xcl=c2x-c1x;
  int ycl=c2y-c1y;

  int n,t;

  DWORD offs; long addr;
  WORD register PalWrite=(PalIndex << 8) | PalIndex;
  

  int dy=c1y-c0y;
  int dx;

  for (y=0; y<dy; y++)
  {
   x1=MulDiv(y,xal,yal);
   x2=MulDiv(y,xbl,ybl);

   if (x1>x2) {t=x2; x2=x1; x1=t;};
   dx=x2-x1;

   offs=c0x+x1+offset[y+c0y];
   if (LOWORD(offs)<=(65535-dx))
   {
    WORD register FAR *ad;
    ad=(WORD FAR *)MAKELONG(LOWORD(offs), HIWORD(bits) +
		      HIWORD(offs) * FP_OFF(__ahIncr));

    n=dx/2;
    for (int register d=0; d<n; d++) ad[d]=PalWrite;
    if (dx % 2) (BYTE)ad[n]=BYTE(PalIndex);
   }
   else
   for (x=c0x+x1; x<=c0x+x2; x++)
   {
    offs=x+offset[y+c0y];

    addr=MAKELONG(LOWORD(offs), HIWORD(bits) +
		      HIWORD(offs) * FP_OFF(__ahIncr));

    *(BYTE FAR *)addr=BYTE(PalIndex);
   }
  }

  for (y=c1y; y<c2y; y++)
  {
   x1=c0x+MulDiv(y-c0y,xal,yal);
   x2=c1x+MulDiv(y-c1y,xcl,ycl);

   if (x1>x2) {t=x2; x2=x1; x1=t;};

   offs=x1+offset[y];
   if (LOWORD(offs)<=(65535-x2+x1))
   {
    WORD register FAR *ad;
    ad=(WORD FAR *)MAKELONG(LOWORD(offs), HIWORD(bits) +
		      HIWORD(offs) * FP_OFF(__ahIncr));

    n=(x2-x1)/2;
    for (int register d=0; d<n; d++) ad[d]=PalWrite;
    if ((x2-x1) % 2) (BYTE)ad[n]=BYTE(PalIndex);
   }
   else
   for (x=x1; x<=x2; x++)
   {
    offs=x+offset[y];

    addr=MAKELONG(LOWORD(offs), HIWORD(bits) +
		      HIWORD(offs) * FP_OFF(__ahIncr));

    *(BYTE FAR *)addr=BYTE(PalIndex);
   }
  }
 }
}


*/
